-- gfight

do
	local lset = l3dset.getnormal()
	local r,g,b = 0.8,0.8,0.8
	--l3dset.getdefaultview(l3dset.getbackground()):backgroundcolor(r,g,b,1)
	layer = lset:layer0()
	sunl3d = l3dlight.new("sun",layer)
	sunl3d:makesun()
	sun = actornode.new("")
	sun:pos(0,100,0)
	sunl3d:diffuse(1.2,1.2,1.2,0)
	local amb = 0.8
	sunl3d:ambient(amb*1.6,amb,amb,0)
	sunl3d:linkinterface(sun)
end


globalspace = dspacehash.new()

Mass = {
	list = {},
	space = dspacehash.new()
}
globalspace:add(Mass.space)

sunsys = particlesys.load("sunsys.prt")
function Mass.new (type, x,z, r,mass)
	local self = {
		mass = mass,
		actor = actornode.new("mass"),
		geom = dgeomsphere.new(r),
	}
	setmetatable(self,{__index = Mass})
	if (type=="sun") then
		self.l3d = l3dpemitter.new("sunmass",sunsys,l3dset.gettop():layer0())
		self.l3d:linkinterface(self.actor)
		self.l3d:start(0)
	else
		self.l3d = l3dprimitive.newsphere("mass",r)
		self.l3d:linkinterface(self.actor)
	end
	self.actor:pos(x,160,z)
	self.geom:pos(x,200,z)
	self.l3d:color(1,1,0,1)
	self.geom.mass = true
	self.space:add(self.geom)
	table.insert(self.list,self)
	return self
end

function Mass:think ()
	local pos = {self.actor:pos()}
	local function grav(sat)
		local dif = ExtMath.v3dif(pos,{sat.body:pos()})
		local f = 1/ExtMath.v3sqlen(dif)*self.mass
		local vx,vy,vz = sat.body:velocity()
		sat.body:velocity(dif[1]*f+vx,vy,dif[3]*f+vz)
	end
	for i,sat in ipairs(Satellite.list) do
		grav(sat)
	end
	for i,p in ipairs(Projectile.list) do
		grav(p)
	end
end

Projectile = {
	list = {},
	space = dspacehash.new()
}

projsys = particlesys.load("projectile.prt")
globalspace:add(Projectile.space)
--Projectile.space:collidetest(false)

function Projectile.new (x,z,vx,vz, owner)
	local self = {
		actor = actornode.new("proj",x,200,z),
		l3d = l3dprimitive.newsphere("proj",0.4,0.4,0.4),
		trail = l3dtrail.new("proj",50),
		geom = dgeomsphere.new(0.5),
		body = dbody.new(),
	}
	setmetatable(self,{__index = Projectile})
	self.actor:link(self.body)
	self.l3d:linkinterface(self.actor)
	self.l3d:rfLitSun(true)
	self.trail:linkinterface(self.actor)
	self.trail:spawnstep(20)
	self.trail:facecamera(true)
	self.trail:color(0,1,0,1,1)
	self.trail:color(1,0,0.5,0.5,1)
	self.trail:color(2,0,0,0,1)
	self.trail:size(0,0.2)
	self.trail:size(1,0.1)
	self.trail:size(2,0)
	self.trail:rfBlend(true)
	self.trail:rsBlendmode(blendmode.add())
	self.trail:rfNodepthmask(true)
	
	self.body:masssphere(1.5,0.5,true)
	self.geom:body(self.body)
	self.space:add(self.geom)
	self.body:velocity(vx,0,vz)
	self.body:pos(x,200,z)
	self.body:lock(2,200)
	self.id = #Projectile.list + 1
	Projectile.list[self.id] = self
	TimerTask.new(
		function () 
			if (self.deleted) then return end
			self.brake = true self.trail:color(0,1,0,0,1) 
		end, 
		8000)
	return self
end

function Projectile:think ()
	if (self.deleted) then return end
	local vx,vy,vz = self.body:velocity()
	local p = {self.body:pos()}
	p[2] = 0
	local sqlen = ExtMath.v3sqlen(p)
	self.sqdist = sqlen
	if (sqlen>130*130) then print (sqlen) self:delete() return true end
	
	if (self.brake) then 
		self.body:velocity(vx*0.996,vy*0.996,vz*0.996)
	end
	for i=0,self.body:jointcount()-1 do
		local j = self.body:joint(i)
		local g1,g2 = j:geom1(),j:geom2()
		if (g1.mass or g2.mass) then
			self:delete()
			return true
		end
	end
end

function Projectile:deleteall ()
	while (#Projectile.list>0) do
		table.remove(Projectile.list):delete()
	end
end

function Projectile:delete()
	if (self.deleted) then return end
	local p = table.remove(Projectile.list)
	if (p~=self) then
		Projectile.list[self.id] = p
		p.id = self.id
	end
	self.deleted = true
	self.actor:delete()
	self.l3d:delete()
	self.body:delete()
	self.geom:delete()
	self.trail:delete()
end

Satellite = {
	list = {},
	space = dspacehash.new()
}

globalspace:add(Satellite.space)

function Satellite:deleteall ()
	while (#Satellite.list>0) do
		table.remove(Satellilte.list):delete()
	end
end

local colors = {
	{1,0,0,1}, {0,1,0,1}, {0,0,1,1}, {1,1,0,1}, {0,1,1,1}, {1,0,1,1}
}

function Satellite:resettrails ()
	for i,v in ipairs(self.trails) do v:reset() end	
end

ship = model.load("ships/protector-mk1.f3d",false,true)

function Satellite.new (steerfn,x,z,vx,vz)
	local self = {
		steer = steerfn,
		id = #Satellite.list + 1,
		actor = actornode.new("sat"),
		body = dbody.new(),
		geom = dgeombox.new(4,1,4),
		trails = {
			l3dtrail.new("sattrail",400),
			l3dtrail.new("sattrail",400),
		},
		l3d = l3dmodel.new("ship",ship,false,false)
--		l3dsp = l3dprimitive.newsphere("sat",0.75),
--		l3d = l3dprimitive.newcube("sat",4,0.5,0.5)
	}
	setmetatable(self,{__index = Satellite})
	Satellite.list[self.id] = self

	self.l3d:color(unpack(colors[self.id]))	
	self.l3d:rfLitSun(true)
	
	--self.l3dsp:color(unpack(colors[self.id]))	
	--self.l3dsp:rfLitSun(true)
	self.l3d:linkinterface(self.actor)
	self.l3d:uselocal(true)
	self.l3d:localrotdeg(90,180,0)
	self.l3d:localpos(0,0,1)
	self.l3d:renderscale(0.15,0.15,0.15)
	--self.l3dsp:linkinterface(self.actor)
	--self.l3dsp:uselocal(true)
	--self.l3dsp:localpos(0,0,0.5)
	for i,trail in ipairs(self.trails) do 
		trail:linkinterface(self.actor)
		trail:rfNocull(true)
		trail:color(0,0.5,0.5,1,1)
		trail:color(1,0,0.5,0.5,1)
		trail:color(2,0,0,0,1)
		trail:rfBlend(true)
		trail:rfNodepthmask(true)
		trail:rsBlendmode(blendmode.add())
		trail:size(0,0.3)
		trail:size(1,0.15)
		trail:size(2,0)
	end
	self.actor:link(self.body)
	self.geom:body(self.body)
	self.body:pos(x,200,z)
	self.body:velocity(vx,0,vz)
	self.body:lock(2,200)
	self.reload = 0
	
	Satellite.space:add(self.geom)
	self.body:massbox(1,4,1,1,true)
	
	return self
end

firesys = particlesys.load("explofire.prt")

function Satellite:think () 
	local fx,fy,fire,nosteer = self:steer(self.reload>0)
	
	if (self.trailreset) then
		self:resettrails()
		self.trailreset = false
	end

	local x,y,z = self.body:pos()
	local w,h = 165,130
	local nx,nz = (x+w/2)%w-w/2, (z+h/2)%h-h/2
	self.body:pos(nx,y,nz)
	self.trailreset = nx*x<0 or nz*z<0
	if (self.trailreset) then self:resettrails() end
	
	local vx,vy,vz = self.body:velocity()
	local norm,len = ExtMath.v3normalize({vx,vy,vz})
	len = math.min(10,len)
	self.body:velocity(norm[1]*len,norm[2]*len,norm[3]*len)
	local ax,ay,az = self.body:angvel()
	--self.body:angvel(math.min(1,ax),math.min(1,ay),math.min(1,az))
	
	if not self.doomed then
		for i = 0, self.body:jointcount()-1 do
			local j = self.body:joint(i)
			if j.geom1 and j:geom1().mass or j.geom2 and j:geom2().mass then
				local l3d = l3dpemitter.new("fire",firesys)
				l3d:linkinterface(self.actor)
				l3d:start(0)
				l3d:uselocal(true)
				l3d:localpos(0,-3,0)
				self.doomed = true
				DebugOutput.set(self.id+1,"Player "..self.id.." ... lost")
				TimerTask.new(
					function () 
						l3d:stop(0,true)
						TimerTask.new(function() l3d:delete() end, 1000)
						self.doomed = false 
						local ang,dist = math.random()*math.pi*2, math.random()*20+20
						local s,c = math.sin(ang),math.cos(ang)
						self.body:pos(s*dist,200,c*dist)
						self.body:velocity(c*0.3,0,-s*0.3)
						for i,v in ipairs(self.trails) do v:reset() end
						DebugOutput.set(self.id+1,"")
					end,2500)
				return
			end
		end
	else return end
	
	if (fire and self.reload>0) then 
		self.reload = -20
		local x,y,z = self.body:local2global(0,0,2)
		local vx,vy,vz = self.body:local2global(0,0,2.4)
		local vvx,vvy,vvz = self.body:velocity()
		Projectile.new(x,z,vx-x+vvx,vz-z+vvz,self)
	end
	self.reload = self.reload + 1
	
	local ax,ay,az = self.body:angvel()
	self.body:angvel(ax*0.9,ay*0.9,az*0.9)
	local x,y,z = self.body:local2global(2,-0.5,0)
	self.trails[1]:spawn(x,y,z,0,1,0)
	local x,y,z = self.body:local2global(-2,-0.5,0)
	self.trails[2]:spawn(x,y,z,0,1,0)
	
	self.body:addrelforce(0,0,fy*0.01)
	self.body:addtorque(0,fx*0.01,0)
	--self.body:addrelforce(0,0.0001,0,0,1,0)
	local a = {self.body:rotaxis()}
	local b = {0,1,0}
	local c = ExtMath.v3cross(a,b)
	--print(("%.2f "):rep(9):format(a[1],a[2],a[3],b[1],b[2],b[3],c[1],c[2],c[3]))
	--print(("%.2f "):rep(9):format(self.body:rotaxis()))
	self.body:rotaxis(a[1],a[2],a[3],b[1],b[2],b[3],c[1],c[2],c[3])
end

function Satellite:delete()
	if (self.deleted) then return end
	local s = table.remove(Satellite.list)
	if (s~=self) then
		Satellite.list[self.id] = s
		s.id = self.id
	end
	self.deleted = true
	self.actor:delete()
	self.l3d:delete()
	--self.l3dsp:delete()
	self.body:delete()
	self.geom:delete()
	for i,v in ipairs(self.trails) do 
		v:delete()
	end
end

server = socket.udp()
if (server) then
	server:settimeout(0)
	server:setsockname("*",18000)
end
svrpartner = {}
svrsteer = {}
timeout = {}
clientsats = {}

clientpartner = nil

function gamethink () 
	dworld.collidetest(globalspace)
	dworld.makecontacts()
	
	
	for i,v in ipairs(Satellite.list) do 
		v:think()
	end
	for i,v in ipairs(Projectile.list) do
		local err,msg = pcall(v.think,v)
		if (not err) then print(msg) end
	end
	
	dworld.step(1)
end

function initsvrgame ()
	if (player) then return end
	player = Satellite.new(
		function (self,canshoot) 
			if (client) then return 0,0,false end
			local fx,fy = 
				(Keyboard.isKeyDown(Keyboard.keycode.LEFT) and -1 or 0) + 
					(Keyboard.isKeyDown(Keyboard.keycode.RIGHT) and 1 or 0),
			 	(Keyboard.isKeyDown(Keyboard.keycode.UP)	and 1 or 0) + 
			 		(Keyboard.isKeyDown(Keyboard.keycode.DOWN) and -1 or 0)
			 local fire = Keyboard.isKeyDown("LCTRL") and canshoot
			 return fx,fy,fire
		end, 0,40,0.25,0)
end

initsvrgame ()

DebugOutput.enable(true)
	
function serverthink ()
	if (client or not server) then return end
	
	while (true) do
		local data,ip,port = server:receivefrom()
		if (not data) then break end
		local id = ("%s:%d"):format(ip,port)
		if (data == "hello") then -- connectting - ignore for now
			initsvrgame()
			table.insert(svrpartner,{ip=ip,port=port})
			svrsteer[id] = {0,0,false}
			clientsats[id] = {Satellite.new(
				function (self) return unpack(svrsteer[id]) end, 0,-40,-0.25,0),ip = ip,port = port}
		elseif (svrsteer[id]) then
			timeout[id] = system.time()
			svrsteer[id] = {
				tonumber(data:sub(1,2)) or 0,
				tonumber(data:sub(3,4)) or 0,
				tonumber(data:sub(5,5))==1}
			--print(ip,unpack(svrsteer[ip]))
		end
	end
	
--	if (#svrpartner==0) then return end
	
	for i,v in ipairs(Mass.list) do
		v:think()
	end
	gamethink()
	
	local packet = {}
	
	for i,ip in ipairs(svrpartner) do
		--table.insert(packet,("C%s:%s;"):format(ip.ip,ip.port))
	end
	
	table.insert(packet,string.char2bin(#Satellite.list))
	local cm = function (var) return string.short2bin(math.floor(var*128+0.5)) end
	for i,self in ipairs(Satellite.list) do
		local x,y,z = self.body:pos()
		local ax,ay,az = self.body:rotaxis()
		local satstr = cm(x)..cm(z)..cm(ax)..cm(az)
		--print("svr",string.byte(satstr),string.byte(satstr,1),string.byte(satstr,2))
		table.insert(packet,satstr)
		--table.insert(packet,("S%d,%f,%f,%f,%f;"):format(i,x,z,ax,az))
	end
	table.insert(packet,string.short2bin(#Projectile.list))
	for i,self in ipairs(Projectile.list) do
		--if (self.sqdist <80*80) then
			local x,y,z = self.body:pos()
			local x,z = math.max(-300,math.min(x,300)),math.max(-300,math.min(z,300))
			--local vx,vy,vz = self.body:velocity()
			table.insert(packet,cm(x)..cm(z))
			--table.insert(packet,("P%d,%f,%f,%f,%f;"):format(i,x,z,vx,vz))
		--end
	end
	
	packet = table.concat(packet)
	DebugOutput.set(1,("packet: %d"):format(#packet))
	
	for i, v in ipairs(svrpartner) do 
		server:sendto(packet,v.ip,v.port)
	end
	
	local t = system.time()
	for id,v in pairs(timeout) do
		if (t-v>2000) then
			clientsats[id][1]:delete()
			
			for i,p in ipairs(svrpartner) do
				local v = clientsats[id]
				print(p,v)
				if (p.ip == v.ip and p.port == v.port) then
					table.remove(svrpartner,i)
					break
				end
			end
			timeout[id] = nil
			clientsats[id] = nil
			svrsteer[id] = nil
		end
	end
end

function clientthink ()
	if not client then return end
	local fx,fy = 
			(Keyboard.isKeyDown(Keyboard.keycode.LEFT) and -1 or 0) + 
				(Keyboard.isKeyDown(Keyboard.keycode.RIGHT) and 1 or 0),
		 	(Keyboard.isKeyDown(Keyboard.keycode.UP)	and 1 or 0) + 
		 		(Keyboard.isKeyDown(Keyboard.keycode.DOWN) and -1 or 0)
	local fire = Keyboard.isKeyDown("LCTRL") and 1 or 0
	client:send(("%+d%+d%d"):format(fx,fy,fire))

	local lastdata
	while (true) do
		local data = client:receive()
		if (not data) then break end
		lastdata = data
	end
	local data = lastdata
	if not data then return end
	local maxS = data:sub(1,1):bin2char()
	
	for i=1,maxS do
		local satstr = data:sub((i-1)*8+2,i*8+2)
		local id = i
		local cut = function (a) a=a*2+1 return satstr:sub(a,a+1):bin2short()/128 end
		local x,y,dirx,dirz = cut(0),cut(1),cut(2),cut(3)
		--print(string.byte(satstr),string.byte(satstr,1),string.byte(satstr,2))
		--if true then break end
		--print(x,y,dirx,dirz)
	--for satstr in data:gfind("S(.-);") do
	--	local id,x,y,dirx,dirz = satstr:match("(%d+),([%d.e%-%+]-),([%d.e%-%+]-),([%d.e%-%+]-),([%d.e%-%+]-)$")
		if (not id) then print("error") 
		else 
			
			if not Satellite.list[id] then Satellite.new(function () return 0,0,false end, 0,-40,-0.25,0) end
			--maxS = math.max(id,maxS)
			local sat = Satellite.list[id]
			--print(string.format("%.2f %.2f %.2f\n%.2f %.2f %.2f",x,200,y,sat.body:pos()))
			--local X,Y,Z = sat.body:pos()
			sat.body:pos(x,200,y)
			
			local dir = {dirx,0,dirz}
			local up = {0,1,0}
			local right = ExtMath.v3cross(dir,up)
			
			sat.body:rotaxis(dir[1],dir[2],dir[3],up[1],up[2],up[3],right[1],right[2],right[3])
		end
		--print(unpack(right))
		--print(ax,ay,az)
		
		--print(("%s %d %d %d"):format(satstr,id or 0,x or 0,y or 0,dir or 0))
	end
	while (Satellite.list[maxS+1]) do Satellite.list[maxS+1]:delete() end
	
	local offs = maxS*8+2
	local maxP = data:sub(offs,offs+2):bin2short()
	for i=1,maxP do 
		local pstr = data:sub(offs+(i-1)*4+2,offs+i*4+2)
		local id = i
		local cut = function (a) a=a*2+1 return pstr:sub(a,a+1):bin2short()/128 end
		local x,z = cut(0),cut(1)
		while (not Projectile.list[id]) do Projectile.new(0,0,0,0,nil) end
		local dist = ExtMath.v3sqdist({Projectile.list[id].body:pos()},{x,200,z})
		if (dist>25) then Projectile.list[id].trail:reset() end
		Projectile.list[id].body:pos(x,200,z)
	end
	while (Projectile.list[maxP+1]) do Projectile.list[maxP+1]:delete() end
	
	gamethink()
end

function con (ip)
	client = socket.udp()
	client:settimeout(0)
	--client:setsockname("*",17999)
	client:setpeername(ip,18000)
	
	client:send("hello")
end

--con("127.0.0.1",true)



function coneip() con("192.168.0.163") end
function concip() con("192.168.0.191") end
function conself() con("127.0.0.1") end

Mass.new("sun",0,0,5,1)	
Timer.set("gfight", 
	function (t) 
		serverthink()
		clientthink()
		
		--gamethink()
	end, 15,5)